
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <asm/atari_SLM.h>


#define	SENDMAIL_PATH	"/usr/lib/sendmail"

char	*ProgName;		/* name of this program */
char	HostName[256];	/* name of this host */
char	*AcctFileName;	/* name of accounting file */
FILE	*AcctFile;		/* accouting file pointer */
char	*UserName;		/* name of user that prints */
char	CurrStatus[80];	/* string describing current status of printer */
int		PagesPrinted;	/* # of pages printed */
time_t	StartTime = 0,	/* start and end time of printing, or 0 */
		EndTime = 0;

#define	BORDER			75	/* non-printable 1/4'' at each border */


#define	ERR_WAIT		1, 0, 0
#define	ERR_MAIL		0, 1, 0
#define	ERR_MAIL_WAIT	1, 1, 0
#define	ERR_MAIL_ABORT	0, 1, 1

struct err_action {
	char		*text;
	unsigned	wait : 1;
	unsigned	mail : 1;
	unsigned	abort : 1;
} ErrAction[] = {
	{ /* 0x02 */ "ornery printer",			ERR_MAIL_ABORT },
	{ /* 0x03 */ "toner empty",				ERR_MAIL_ABORT },
	{ /* 0x04 */ "warming up",				ERR_WAIT },
	{ /* 0x05 */ "paper empty",				ERR_MAIL_WAIT },
	{ /* 0x06 */ "drum empty",				ERR_MAIL_ABORT },
	{ /* 0x07 */ "input jam",				ERR_MAIL_WAIT },
	{ /* 0x08 */ "through jam",				ERR_MAIL_WAIT },
	{ /* 0x09 */ "output jam",				ERR_MAIL_WAIT },
	{ /* 0x0a */ "cover open",				ERR_MAIL_WAIT },
	{ /* 0x0b */ "fuser malfunction",		ERR_MAIL_ABORT },
	{ /* 0x0c */ "imager malfunction",		ERR_MAIL_ABORT },
	{ /* 0x0d */ "motor malfunction",		ERR_MAIL_ABORT },
	{ /* 0x0e */ "video malfunction",		ERR_MAIL_ABORT },
	{ /* 0x10 */ "printer system timeout",	ERR_MAIL_ABORT },
	{ /* 0x12 */ "invalid operation code",	ERR_MAIL_ABORT },
	{ /* 0x15 */ "invalid device number",	ERR_MAIL_ABORT },
	{ /* 0x1a */ "invalid parameter list",	ERR_MAIL_ABORT },
	{ /* 0x1b */ "ACSI timeout",			ERR_MAIL_ABORT },
	{ /* 0x1c */ "not all printed",			ERR_MAIL }
};

#define	N_ERR_ACTION	(sizeof(ErrAction)/sizeof(*ErrAction))

struct err_action DefaultErrAction = {
	"unknown error", ERR_MAIL_ABORT
};


/***************************** Prototypes *****************************/

int main( int argc, char *argv[] );
void parse_args( int argc, char *argv[] );
void usage( const char *format, ... );
int print_page( int fd, int w, int h );
int readline( int fd, char *buf, int cnt, int first_line );
void send_mail( const char *subject, const char *format, ... );
void error( const char *format, ... );
void get_papersize( int *w, int *h );
int get_status( void );
void set_status( const char *str );
struct err_action *get_err_action( void );
void do_err_action( struct err_action *ea );
void write_acct( void );
void set_signals( void );
void my_sig_handler( int signr );

/************************* End of Prototypes **************************/



int main( int argc, char *argv[] )

{	int					stat, w, h, pipefd[2], pid;
	char				geom[40];
	struct err_action	*ea;
	
	if ((ProgName = strrchr( argv[0], '/' )))
		++ProgName;
	else
		ProgName = argv[0];

	if (gethostname( HostName, sizeof(HostName) )) {
		perror( "gethostname" );
		exit( 1 );
	}
	
	parse_args( argc, argv );

	if (AcctFileName) {
		if (!(AcctFile = fopen( AcctFileName, "a" )))
			error( "Cannot open accounting file %s: %s\n",
				   AcctFileName, strerror(errno) );
	}
	atexit( write_acct );
	set_signals();
	/* reset the SLM buffer (in case there are some data left over...) */
	if (ioctl( 1, SLMIORESET, NULL ))
		error( "Cannot reset SLM: %s\n", strerror(errno) );
	
	if ((stat = get_status())) {
		ea = get_err_action();
		do_err_action( ea );
	}
	get_papersize( &w, &h );
	w += 2*BORDER;
	h += 2*BORDER;
	PagesPrinted = 0;
	time( &StartTime );

	/* fork off GhostScript process reading from out stdin */

	if (pipe( pipefd ))
		error( "%s: pipe: %s\n", strerror(errno) );
		
	set_status( "Formatting" );
	if ((pid = fork()) == 0) {
		/* child */

		/* redirect stdout to the pipe */
		close( 1 );
		dup( pipefd[1] );
		close( pipefd[1] );
		/* close reading end of pipe */
		close( pipefd[0] );

		/* build command line and start gs */
		sprintf( geom, "-g%dx%d", w, h );
		execlp( "gs", "gs", "-q", "-dNOPAUSE", "-dSAFER", "-sOutputFile=-",
				"-sDEVICE=bit", "-r300", geom, "-", "quit.ps", NULL );
		error( "%s: cannot exec gs: %s\n", ProgName, strerror(errno) );
	}
	else if (pid < 0)
		/* error */
		error( "fork GhostScript: %s\n", strerror(errno) );

	/* parent */

	/* close stdin (that's for gs) and use stderr there */
	close( 0 );
	dup( 2 );
	/* close writing end of pipe */
	close( pipefd[1] );
	
	for(;;) {
		if (print_page( pipefd[0], w, h )) break;
		++PagesPrinted;
		if ((stat = get_status())) {
			ea = get_err_action();
			do_err_action( ea );
		}
	}

	time( &EndTime );
	if ((stat = get_status())) {
		ea = get_err_action();
		do_err_action( ea );
	}

	/* get gs' exit status */
	wait( &stat );
	if (stat)
		/* just report */
		fprintf( stderr, "%s: gs returned status %d\n", ProgName, stat );

	return( 0 );
}


#define	SETUP_ARG(x)											\
	do {														\
		if (p[1])												\
			q = p+1;											\
		else {													\
			if (++i == argc)									\
				usage( "-" x " option missing argument\n" );	\
			q = argv[i];										\
		}														\
	} while( 0 )

void parse_args( int argc, char *argv[] )

{	int		i;
	char	*p, *q, *name, *host;
	
	AcctFileName = NULL;
	UserName = NULL;
	name = NULL;
	host = NULL;
	
	for( i = 1; i < argc; ++i ) {
		p = argv[i];
		if (*p == '-') {
			switch( *++p ) {
			  case 'w':
				SETUP_ARG("w");
				/* ignore */
				break;
			  case 'l':
				SETUP_ARG("l");
				/* ignore */
				break;
			  case 'i':
				SETUP_ARG("i");
				/* ignore */
				break;
			  case 'x':
				SETUP_ARG("x");
				/* ignore */
				break;
			  case 'y':
				SETUP_ARG("y");
				/* ignore */
				break;
			  case 'c':
				/* ignore */
				break;
			  case 'n':
				SETUP_ARG("n");
				name = q;
				break;
			  case 'h':
				SETUP_ARG("h");
				host = q;
				break;
			  default:
				usage( "Bad option\n" );
			}
		}
		else {
			AcctFileName = argv[i++];
			break;
		}
	}
	if (i != argc) usage( "Too many arguments\n" );

	if (name) {
		if (!host)
			host = HostName;
		UserName = (char *)malloc( strlen(name)+strlen(host)+2 );
		strcpy( UserName, name );
		strcat( UserName, "@" );
		strcat( UserName, host );
	}
}


void usage( const char *format, ... )

{	va_list	val;
	
	va_start( val, format );
	vfprintf( stderr, format, val );
	va_end( val );
	
	fprintf( stderr, "Usage: %s [-n name] [-h host] [acctfile]\n", ProgName );
	exit( 1 );
}


int print_page( int fd, int w, int h )

{	int				i, j, n;
	int				byte_width = (w + 7)/8;
	unsigned char	buffer[byte_width];
	int				inner_start = BORDER >> 3,
					inner_width = (w-2*BORDER)/8,
					shift = BORDER & 7;
	unsigned char	obuffer[inner_width];
	unsigned char	*p, *q;

	set_status( "Formatting" );
	
	/* first skip 'Border' lines of top margin */
	for( i = 0; i < BORDER; ++i )
		if (readline( fd, buffer, byte_width, i == 0 ))
			return( 1 );

	/* then process h-2*BORDER lines by horizontal cutting 'BORDER' bits
	 * at both sides */
	n = h-2*BORDER;
	for( i = 0; i < n; ++i ) {
		readline( fd, buffer, byte_width, 0 );
		for( j = 0, p = buffer+inner_start, q = obuffer; j < inner_width;
			 ++j, ++p )
			*q++ = (p[0] << shift) | (p[1] >> (8-shift));
		if (i != n-1)
			write( 1, obuffer, inner_width );
	}
	
	/* And skip 'BORDER' lines at the bottom */
	for( i = 0; i < BORDER; ++i )
		readline( fd, buffer, byte_width, 0 );

	/* writing the last line starts the printing */
	set_status( "Printing" );
	write( 1, obuffer, inner_width );

	return( 0 );
}


int readline( int fd, char *buf, int cnt, int first_line )

{	int		n, readed = 0;

	while( readed < cnt ) {
		n = read( fd, buf, cnt-readed );
		if (n < 0)
			error( "reading from gs pipe: %s\n", strerror(errno) );
		else if (n == 0) {
			if (!first_line)
				error( "unexpected end of data while reading from gs pipe\n" );
			else
				return( 1 );
		}
		readed += n;
	}
	return( 0 );
}


void send_mail( const char *subject, const char *format, ... )

{	va_list	val;
	char	cmdline[1024];
	FILE	*pipe;
	
	if (!UserName)
		return;

	strcpy( cmdline, SENDMAIL_PATH " " );
	strcat( cmdline, UserName );
	if (!(pipe = popen( cmdline, "w" ))) {
		fprintf( stderr, "%s: Cannot open pipe to " SENDMAIL_PATH ": %s\n",
				 ProgName, strerror(errno) );
		return;
	}

	fprintf( pipe, "From: PRINTER-DAEMON\n" );
	fprintf( pipe, "Subject: %s\n", subject );
	fprintf( pipe, "Reply-To: <lpadm@%s>\n\n", HostName );
	
	va_start( val, format );
	vfprintf( pipe, format, val );
	va_end( val );
	
	pclose( pipe );
}


void error( const char *format, ... )

{	va_list	val;
	char	buf[1024];
	
	va_start( val, format );
	vsprintf( buf, format, val );
	va_end( val );

	fprintf( stderr, "%s: %s", ProgName, buf );
	send_mail( "printer error", buf );
	
	exit( 1 );
}


void get_papersize( int *w, int *h )

{	struct SLM_paper_size	psize;
	
	if (ioctl( fileno(stdout), SLMIOGPSIZE, &psize ))
		error( "Cannot get SLM's current paper size: %s\n", strerror(errno) );

	*w = psize.width;
	*h = psize.height;
}


int get_status( void )

{	struct SLM_status	pstat;
	
	if (ioctl( fileno(stdout), SLMIOGSTAT, &pstat ) < 0)
		error( "Cannot get SLM's current status: %s\n", strerror(errno) );

	strcpy( CurrStatus, pstat.str );
	set_status( CurrStatus );
	return( pstat.stat );
}


void set_status( const char *str )

{	FILE	*f;

	if (!(f = fopen( "status", "w" )))
		return;
	fprintf( f, "%s\n", str );
	fclose( f );
}


struct err_action *get_err_action( void )

{	int		i;

	for( i = 0; i < N_ERR_ACTION; ++i )
		if (!strcasecmp( CurrStatus, ErrAction[i].text ))
			break;

	if (i < N_ERR_ACTION)
		return( &ErrAction[i] );
	else
		return( &DefaultErrAction );
}


void do_err_action( struct err_action *ea )

{	struct err_action	*ea2;
	int					stat;
	
  restart:
	if (ea->mail) {
		send_mail( "printer error",
				   "The following printer error occured:\n"
				   "\t%s\n"
				   "%s\n\n"
				   "Your printer subsystem\n\n",
				   CurrStatus,
				   ea->wait  ? "Please remove this condition and your "
							   "print job will continue." :
				   ea->abort ? "Your print job had to be aborted" :
							   "Your print job was continued, but may "
							   "include bad pages" );
	}

	if (ea->abort)
		exit( 1 );

	if (ea->wait) {
		for(;;) {
			sleep( 5 );
			stat = get_status();
			if (stat == 0)
				/* no more error -> waiting finished */
				break;
			ea2 = get_err_action();
			if (ea2 != ea)
				/* another error condition -> do appropriate actions */
				goto restart;
		}
	}
}


void write_acct( void )

{	char		buf[256];
	struct tm	*tm;
	
	if (!AcctFileName || !PagesPrinted)
		return;

	if (!EndTime) time( &EndTime );

	tm = localtime( &StartTime );
	strftime( buf, sizeof(buf), "%x %X", tm );
	fprintf( AcctFile, "%s-", buf );
	tm = localtime( &EndTime );
	strftime( buf, sizeof(buf), "%X", tm );
	fprintf( AcctFile, "%s: %d page(s) printed by %s\n",
			 buf, PagesPrinted, UserName );
}


/* catch all signals for writing to the account file, if we're killed
 * by lprm */ 

void set_signals( void )

{
	signal( SIGHUP, my_sig_handler );
	signal( SIGINT, my_sig_handler );
	signal( SIGQUIT, my_sig_handler );
	signal( SIGILL, my_sig_handler );
	signal( SIGTRAP, my_sig_handler );
	signal( SIGABRT, my_sig_handler );
	signal( SIGIOT, my_sig_handler );
	signal( SIGBUS, my_sig_handler );
	signal( SIGFPE, my_sig_handler );
	signal( SIGUSR1, my_sig_handler );
	signal( SIGSEGV, my_sig_handler );
	signal( SIGUSR2, my_sig_handler );
	signal( SIGPIPE, my_sig_handler );
	signal( SIGALRM, my_sig_handler );
	signal( SIGTERM, my_sig_handler );
	signal( SIGSTKFLT, my_sig_handler );
	signal( SIGXCPU, my_sig_handler );
	signal( SIGXFSZ, my_sig_handler );
	signal( SIGVTALRM, my_sig_handler );
}


void my_sig_handler( int signr )

{
	fprintf( stderr, "%s: Killed by signal %s\n",
			 ProgName, sys_siglist[signr] );
	get_status();
	exit( 1 );
}
